/* BASIC file system interface. */
/* #includes */ /*{{{C}}}*//*{{{*/
#include "config.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#ifdef HAVE_TERMCAP_H
#include <termcap.h>
#endif
#ifdef HAVE_CURSES_H
#include <curses.h>
#endif
#include <unistd.h>

#include "fs.h"

#ifdef USE_DMALLOC
#include "dmalloc.h"
#endif
/*}}}*/
/* #defines */ /*{{{*/
#define LINEWIDTH 80
#define COLWIDTH  14
/*}}}*/

static struct FileStream **file;
static int capacity;
static int used;
static struct termios origMode,rawMode;

const char *FS_errmsg;
static char FS_errmsgbuf[80];
#ifdef HAVE_TGETENT
static char *term,entrybuf[2048],capbuf[2048],*cap;
static char capcl[]="cl",capcm[]="cm";
static char *cl,*cm;
static int termchannel;
#endif

static void size(int dev) /*{{{*/
{
  if (dev>=capacity)
  {
    int i;

    file=realloc(file,(dev+1)*sizeof(struct FileStream*));
    for (i=capacity; i<=dev; ++i) file[i]=(struct FileStream*)0;
    capacity=dev+1;
  }
}
/*}}}*/
static int opened(int dev, int mode) /*{{{*/
{
  int fd;
  const char *msg;

  if (dev<0 || dev>=capacity || file[dev]==(struct FileStream*)0)
  {
    sprintf(FS_errmsgbuf,"channel #%d not open",dev);
    FS_errmsg=FS_errmsgbuf;
    return -1;
  }
  switch (mode)
  {
    case 0: fd=file[dev]->outfd; msg="writing"; break;
    case 1: fd=file[dev]->infd; msg="reading"; break;
    case 2: fd=file[dev]->randomfd; msg="random access"; break;
    default: assert(0);
  }
  if (fd==-1)
  {
    sprintf(FS_errmsgbuf,"channel #%d not opened for %s",dev,msg);
    FS_errmsg=FS_errmsgbuf;
    return -1;
  }
  else return 0;
}
/*}}}*/
static int refill(int dev) /*{{{*/
{
  struct FileStream *f;
  ssize_t len;

  f=file[dev];
  f->inSize=0;
  len=read(f->infd,f->inBuf,sizeof(f->inBuf));
  if (len<=0)
  {
    f->inCapacity=0;
    FS_errmsg=(len==-1?strerror(errno):(const char*)0);
    return -1;
  }
  else
  {
    f->inCapacity=len;
    return 0;
  }
}
/*}}}*/
static int edit(int chn, int onl) /*{{{*/
{
  struct FileStream *f=file[chn];
  char *buf=f->inBuf;
  char ch;
  int r;

  for (buf=f->inBuf; buf<(f->inBuf+f->inCapacity); ++buf)
  {
    if (*buf>='\0' && *buf<' ')
    {
      FS_putChar(chn,'^');
      FS_putChar(chn,*buf?(*buf+'a'-1):'@');
    }
    else FS_putChar(chn,*buf);
  }
  do
  {
    FS_flush(chn);
    if ((r=read(f->infd,&ch,1))==-1)
    {
      f->inCapacity=0;
      FS_errmsg=strerror(errno);
      return -1;
    }
    else if (r==0 || (f->inCapacity==0 && ch==4))
    {
      FS_errmsg=(char*)0;
      return -1;
    }
    if (ch==rawMode.c_cc[VERASE])
    {
      if (f->inCapacity)
      {
        if (f->inBuf[f->inCapacity-1]>='\0' && f->inBuf[f->inCapacity-1]<' ') FS_putChars(chn,"\b\b  \b\b");
        else FS_putChars(chn,"\b \b");
        --f->inCapacity;
      }
    }
    else if ((f->inCapacity+1)<sizeof(f->inBuf))
    {
      if (ch!='\n')
      {
        if (ch>='\0' && ch<' ')
        {
          FS_putChar(chn,'^');
          FS_putChar(chn,ch?(ch+'a'-1):'@');
        }
        else FS_putChar(chn,ch);
      }
      else if (onl) FS_putChar(chn,'\n');
      f->inBuf[f->inCapacity++]=ch;
    }
  } while (ch!='\n');
  return 0;
}
/*}}}*/
#ifdef HAVE_TGETENT
static int initTerminal(int chn) /*{{{*/
{
  static int init=0;

  if (!init)
  {
    if ((term=getenv("TERM"))==(char*)0)
    {
      FS_errmsg="environment variable TERM is not set";
      return -1;
    }
    switch (tgetent(entrybuf,term))
    {
      case -1:
      {
        FS_errmsg="reading terminal description failed";
        return -1;
      }
      case 0:
      {
        sprintf(FS_errmsgbuf,"unknown terminal type %s",term);
        FS_errmsg=FS_errmsgbuf;
        return -1;
      }
      case 1:
      {
        cl=tgetstr(capcl,&cap);
        cm=tgetstr(capcm,&cap);
        return 0;
      }
    }
  }
  return 0;
}
/*}}}*/
static int outc(int ch) /*{{{*/
{
  struct FileStream *f;

  if (opened(termchannel,0)==-1) return -1;
  f=file[termchannel];
  if (f->outSize+1>=f->outCapacity && FS_flush(termchannel)==-1) return -1;
  f->outBuf[f->outSize++]=ch;
  FS_errmsg=(const char*)0;
  return ch;
}
/*}}}*/
static int cls(int chn) /*{{{*/
{
  if (initTerminal(chn)==-1) return -1;
  termchannel=chn;
  if (cl==(char*)0)
  {
    sprintf(FS_errmsgbuf,"terminal type %s can not clear the screen",term);
    FS_errmsg=FS_errmsgbuf;
    return -1;
  }
#ifdef TPUTS_RETURNS_VOID
  tputs(cl,0,outc);
#else
  if (tputs(cl,0,outc)==-1) return -1;
#endif
  return 0;
}
/*}}}*/
static int locate(int chn, int line, int column) /*{{{*/
{
  if (initTerminal(chn)==-1) return -1;
  termchannel=chn;
  if (cm==(char*)0)
  {
    sprintf(FS_errmsgbuf,"terminal type %s can not position the cursor",term);
    FS_errmsg=FS_errmsgbuf;
    return -1;
  }
#ifdef TPUTS_RETURNS_VOID
  tputs(tgoto(cm,column-1,line-1),0,outc);
#else
  if (tputs(tgoto(cm,column-1,line-1),0,outc)==-1) return -1;
#endif
  return 0;
}
/*}}}*/
#else
static int cls(int chn) /*{{{*/
{
  FS_errmsg="This installation does not support terminal handling";
  return -1;
}
/*}}}*/
static int locate(int chn, int line, int column) /*{{{*/
{
  FS_errmsg="This installation does not support terminal handling";
  return -1;
}
/*}}}*/
#endif

int FS_opendev(int chn, int infd, int outfd) /*{{{*/
{
  size(chn);
  if (file[chn]!=(struct FileStream*)0)
  {
    FS_errmsg="file already open";
    return -1;
  }
  file[chn]=malloc(sizeof(struct FileStream));
  file[chn]->dev=1;
  if ((file[chn]->tty=(infd==0 ? isatty(infd) : 0)))
  {
    if (tcgetattr(infd,&origMode)==-1)
    {
      FS_errmsg=strerror(errno);
      free(file[chn]);
      file[chn]=(struct FileStream*)0;
      return -1;
    }
    rawMode=origMode;
    rawMode.c_lflag&=~(ICANON|ECHO);
    rawMode.c_cc[VMIN]=1;
    rawMode.c_cc[VTIME]=0;
    if (tcsetattr(infd,TCSADRAIN,&rawMode)==-1)
    {
      FS_errmsg=strerror(errno);
      free(file[chn]);
      file[chn]=(struct FileStream*)0;
      return -1;
    }
  }
  file[chn]->infd=infd;
  file[chn]->inSize=0;
  file[chn]->inCapacity=0;
  file[chn]->outfd=outfd;
  file[chn]->outPos=0;
  file[chn]->outLineWidth=LINEWIDTH;
  file[chn]->outColWidth=COLWIDTH;
  file[chn]->outCapacity=sizeof(file[chn]->outBuf);
  file[chn]->outSize=0;
  file[chn]->randomfd=-1;
  FS_errmsg=(const char*)0;
  ++used;
  return 0;
}
/*}}}*/
int FS_openin(const char *name) /*{{{*/
{
  int chn,fd;

  if ((fd=open(name,O_RDONLY))==-1)
  {
    FS_errmsg=strerror(errno);
    return -1;
  }
  for (chn=0; chn<capacity; ++chn) if (file[chn]==(struct FileStream*)0) break;
  size(chn);
  file[chn]=malloc(sizeof(struct FileStream));
  file[chn]->dev=0;
  file[chn]->tty=0;
  file[chn]->infd=fd;
  file[chn]->inSize=0;
  file[chn]->inCapacity=0;
  file[chn]->outfd=-1;
  file[chn]->randomfd=-1;
  FS_errmsg=(const char*)0;
  ++used;
  return chn;
}
/*}}}*/
int FS_openinChn(int chn, const char *name) /*{{{*/
{
  int fd;

  size(chn);
  if (file[chn]!=(struct FileStream*)0)
  {
    FS_errmsg="channel already open";
    return -1;
  }
  if ((fd=open(name,O_RDONLY))==-1)
  {
    FS_errmsg=strerror(errno);
    return -1;
  }
  file[chn]=malloc(sizeof(struct FileStream));
  file[chn]->dev=0;
  file[chn]->tty=0;
  file[chn]->infd=fd;
  file[chn]->inSize=0;
  file[chn]->inCapacity=0;
  file[chn]->outfd=-1;
  file[chn]->randomfd=-1;
  FS_errmsg=(const char*)0;
  ++used;
  return chn;
}
/*}}}*/
int FS_openout(const char *name, int tmp) /*{{{*/
{
  int chn,fd;

  if ((fd=open(name,O_WRONLY|O_TRUNC|O_CREAT|(tmp?O_EXCL:0),(tmp?0600:0666)))==-1)
  {
    FS_errmsg=strerror(errno);
    return -1;
  }
  for (chn=0; chn<capacity; ++chn) if (file[chn]==(struct FileStream*)0) break;
  size(chn);
  file[chn]=malloc(sizeof(struct FileStream));
  file[chn]->dev=0;
  file[chn]->tty=0;
  file[chn]->infd=-1;
  file[chn]->outfd=fd;
  file[chn]->outPos=0;
  file[chn]->outLineWidth=LINEWIDTH;
  file[chn]->outColWidth=COLWIDTH;
  file[chn]->outSize=0;
  file[chn]->outCapacity=sizeof(file[chn]->outBuf);
  file[chn]->randomfd=-1;
  FS_errmsg=(const char*)0;
  ++used;
  return chn;
}
/*}}}*/
int FS_openoutChn(int chn, const char *name) /*{{{*/
{
  int fd;

  size(chn);
  if (file[chn]!=(struct FileStream*)0)
  {
    FS_errmsg="channel already open";
    return -1;
  }
  if ((fd=open(name,O_WRONLY|O_TRUNC|O_CREAT,0666))==-1)
  {
    FS_errmsg=strerror(errno);
    return -1;
  }
  file[chn]=malloc(sizeof(struct FileStream));
  file[chn]->dev=0;
  file[chn]->tty=0;
  file[chn]->infd=-1;
  file[chn]->outfd=fd;
  file[chn]->outPos=0;
  file[chn]->outLineWidth=LINEWIDTH;
  file[chn]->outColWidth=COLWIDTH;
  file[chn]->outSize=0;
  file[chn]->outCapacity=sizeof(file[chn]->outBuf);
  file[chn]->randomfd=-1;
  FS_errmsg=(const char*)0;
  ++used;
  return chn;
}
/*}}}*/
int FS_openrandomChn(int chn, const char *name, int recLength) /*{{{*/
{
  int fd;

  assert(chn>=0);
  assert(name!=(const char*)0);
  assert(recLength>0);
  size(chn);
  if (file[chn]!=(struct FileStream*)0)
  {
    FS_errmsg="channel already open";
    return -1;
  }
  if ((fd=open(name,O_RDWR|O_CREAT,0666))==-1)
  {
    FS_errmsg=strerror(errno);
    return -1;
  }
  file[chn]=malloc(sizeof(struct FileStream));
  file[chn]->dev=0;
  file[chn]->tty=0;
  file[chn]->infd=-1;
  file[chn]->outfd=-1;
  file[chn]->randomfd=fd;
  file[chn]->recLength=recLength;
  file[chn]->recPos=0;
  file[chn]->recBuf=malloc(recLength);
  memset(file[chn]->recBuf,0,recLength);
  StringField_new(&file[chn]->field);
  FS_errmsg=(const char*)0;
  ++used;
  return chn;
}
/*}}}*/
int FS_flush(int dev) /*{{{*/
{
  ssize_t written;
  size_t offset;

  if (file[dev]==(struct FileStream*)0)
  {
    FS_errmsg="channel not open";
    return -1;
  }
  offset=0;
  while (offset<file[dev]->outSize)
  {
    written=write(file[dev]->outfd,file[dev]->outBuf+offset,file[dev]->outSize-offset);
    if (written==-1)
    {
      FS_errmsg=strerror(errno);
      return -1;
    }
    else offset+=written;
  }
  file[dev]->outSize=0;
  FS_errmsg=(const char*)0;
  return 0;
}
/*}}}*/
int FS_close(int dev) /*{{{*/
{
  if (file[dev]==(struct FileStream*)0)
  {
    FS_errmsg="channel not open";
    return -1;
  }
  if (file[dev]->outfd>=0)
  {
    if (FS_flush(dev)) return -1;
    close(file[dev]->outfd);
  }
  if (file[dev]->randomfd>=0)
  {
    StringField_destroy(&file[dev]->field);
    free(file[dev]->recBuf);
    close(file[dev]->randomfd);
  }
  if (file[dev]->tty) tcsetattr(file[dev]->infd,TCSADRAIN,&origMode);
  if (file[dev]->infd>=0) close(file[dev]->infd);
  free(file[dev]);
  file[dev]=(struct FileStream*)0;
  FS_errmsg=(const char*)0;
  if (--used==0)
  {
    free(file);
    capacity=0;
  }
  return 0;
}
/*}}}*/
void FS_shellmode(int dev) /*{{{*/
{
  if (file[dev]->tty) tcsetattr(file[dev]->infd,TCSADRAIN,&origMode);
}
/*}}}*/
void FS_fsmode(int dev) /*{{{*/
{
  if (file[dev]->tty) tcsetattr(file[dev]->infd,TCSADRAIN,&rawMode);
}
/*}}}*/
int FS_put(int chn) /*{{{*/
{
  ssize_t offset,written;

  if (opened(chn,2)==-1) return -1;
  offset=0;
  while (offset<file[chn]->recLength)
  {
    written=write(file[chn]->randomfd,file[chn]->recBuf+offset,file[chn]->recLength-offset);
    if (written==-1)
    {
      FS_errmsg=strerror(errno);
      return -1;
    }
    else offset+=written;
  }
  FS_errmsg=(const char*)0;
  return 0;
}
/*}}}*/
int FS_putChar(int dev, char ch) /*{{{*/
{
  struct FileStream *f;

  if (opened(dev,0)==-1) return -1;
  f=file[dev];
  if (ch=='\n') f->outPos=0;
  if (ch=='\b' && f->outPos) --f->outPos;
  if (f->outSize+2>=f->outCapacity && FS_flush(dev)==-1) return -1;
  if (f->outLineWidth && f->outPos==f->outLineWidth)
  {
    f->outBuf[f->outSize++]='\n';
    f->outPos=0;
  }
  f->outBuf[f->outSize++]=ch;
  if (ch!='\n' && ch!='\b') ++f->outPos;
  FS_errmsg=(const char*)0;
  return 0;
}
/*}}}*/
int FS_putChars(int dev, const char *chars) /*{{{*/
{
  while (*chars) if (FS_putChar(dev,*chars++)==-1) return -1;
  return 0;
}
/*}}}*/
int FS_putString(int dev, const struct String *s) /*{{{*/
{
  size_t len=s->length;
  const char *c=s->character;

  while (len) if (FS_putChar(dev,*c++)==-1) return -1; else --len;
  return 0;
}
/*}}}*/
int FS_nextcol(int dev) /*{{{*/
{
  struct FileStream *f;

  if (opened(dev,0)==-1) return -1;
  f=file[dev];
  if
  (
    f->outPos%f->outColWidth
    && f->outLineWidth
    && ((f->outPos/f->outColWidth+2)*f->outColWidth)>f->outLineWidth
    && FS_putChar(dev,'\n')==-1
  ) return -1;
  while (f->outPos%f->outColWidth) if (FS_putChar(dev,' ')==-1) return -1;
  return 0;
}
/*}}}*/
int FS_tab(int dev, int position) /*{{{*/
{
  struct FileStream *f=file[dev];

  while (f->outPos<(position-1)) if (FS_putChar(dev,' ')==-1) return -1;
  return 0;
}
/*}}}*/
int FS_width(int dev, int width) /*{{{*/
{
  if (opened(dev,0)==-1) return -1;
  if (width<0)
  {
    FS_errmsg="negative width";
    return -1;
  }
  file[dev]->outLineWidth=width;
  return 0;
}
/*}}}*/
int FS_cls(int chn) /*{{{*/
{
  struct FileStream *f;

  if (opened(chn,0)==-1) return -1;
  f=file[chn];
  if (!f->tty)
  {
    FS_errmsg="not a terminal";
    return -1;
  }
  if (cls(chn)==-1) return -1;
  if (FS_flush(chn)==-1) return -1;
  f->outPos=0;
  return 0;
}
/*}}}*/
int FS_locate(int chn, int line, int column) /*{{{*/
{
  struct FileStream *f;

  if (opened(chn,0)==-1) return -1;
  f=file[chn];
  if (!f->tty)
  {
    FS_errmsg="not a terminal";
    return -1;
  }
  if (locate(chn,line,column)==-1) return -1;
  if (FS_flush(chn)==-1) return -1;
  f->outPos=column-1;
  return 0;
}
/*}}}*/
int FS_getChar(int dev) /*{{{*/
{
  struct FileStream *f;

  if (opened(dev,1)==-1) return -1;
  f=file[dev];
  if (f->inSize==f->inCapacity && refill(dev)==-1) return -1;
  FS_errmsg=(const char*)0;
  if (f->inSize+1==f->inCapacity)
  {
    char ch=f->inBuf[f->inSize];

    f->inSize=f->inCapacity=0;
    return ch;
  }
  else return f->inBuf[f->inSize++];
}
/*}}}*/
int FS_get(int chn) /*{{{*/
{
  ssize_t offset,rd;

  if (opened(chn,2)==-1) return -1;
  offset=0;
  while (offset<file[chn]->recLength)
  {
    rd=read(file[chn]->randomfd,file[chn]->recBuf+offset,file[chn]->recLength-offset);
    if (rd==-1)
    {
      FS_errmsg=strerror(errno);
      return -1;
    }
    else offset+=rd;
  }
  FS_errmsg=(const char*)0;
  return 0;
}
/*}}}*/
int FS_inkeyChar(int dev, int ms) /*{{{*/
{
  struct FileStream *f;
  char c;
  ssize_t len;
  struct termios timedread;

  if (opened(dev,1)==-1) return -1;
  f=file[dev];
  if (f->inSize<f->inCapacity) return f->inBuf[f->inSize++];
  if (ms)
  {
  timedread=rawMode;
  timedread.c_cc[VMIN]=0;
  timedread.c_cc[VTIME]=ms/100;
  if (tcsetattr(f->infd,TCSADRAIN,&timedread)==-1)
  {
    FS_errmsg=strerror(errno);
    return -1;
  }
  }
  FS_errmsg=(const char*)0;
  if ((len=read(f->infd,&c,1))==-1 || (ms && tcsetattr(f->infd,TCSADRAIN,&rawMode)==-1))
  {
    FS_errmsg=strerror(errno);
    return -1;
  }
  return (len==1?c:-1);
}
/*}}}*/
int FS_eof(int chn) /*{{{*/
{
  struct FileStream *f;

  if (opened(chn,1)==-1) return -1;
  f=file[chn];
  if (f->inSize==f->inCapacity && refill(chn)==-1) return 1;
  return 0;
}
/*}}}*/
long int FS_lof(int chn) /*{{{*/
{
  struct FileStream *f;
  struct stat buf;

  if (opened(chn,1)==-1) return -1;
  if (fstat(file[chn]->infd!=-1?file[chn]->infd:file[chn]->outfd,&buf)==-1)
  {
    FS_errmsg=strerror(errno);
    return -1;
  }
  return buf.st_size;
}
/*}}}*/
long int FS_recLength(int chn) /*{{{*/
{
  if (opened(chn,2)==-1) return -1;
  return file[chn]->recLength;
}
/*}}}*/
void FS_field(int chn, struct String *s, long int position, long int length) /*{{{*/
{
  assert(file[chn]);
  String_joinField(s,&file[chn]->field,file[chn]->recBuf+position,length);
}
/*}}}*/
int FS_seek(long int chn, long int record) /*{{{*/
{
  if (opened(chn,2)==-1) return -1;
  if (lseek(file[chn]->randomfd,(off_t)record*file[chn]->recLength,SEEK_SET)==-1)
  {
    FS_errmsg=strerror(errno);
    return -1;
  }
  return 0;
}
/*}}}*/
int FS_appendToString(int chn, struct String *s, char stop, int onl) /*{{{*/
{
  size_t new;
  char *n;
  struct FileStream *f=file[chn];
  int c;

  if (f->tty && f->inSize==f->inCapacity)
  {
    if (edit(chn,onl)==-1) return (FS_errmsg ? -1 : 0);
  }
  do
  {
    n=f->inBuf+f->inSize;
    while (1)
    {
      if (n==f->inBuf+f->inCapacity) break;
      c=*n++;
      if (c=='\n' || c==stop) break;
    }
    new=n-(f->inBuf+f->inSize);
    if (new)
    {
      size_t offset=s->length;

      String_size(s,offset+new);
      memcpy(s->character+offset,f->inBuf+f->inSize,new);
      f->inSize+=new;
      if (*(n-1)=='\n' || *(n-1)==stop)
      {
        if (f->inSize==f->inCapacity) f->inSize=f->inCapacity=0;
        return 0;
      }
    }
    if ((c=FS_getChar(chn))>=0) String_appendChar(s,c);
    if (c=='\n' || c==stop) return 0;
  } while (c!=-1);
  return (FS_errmsg ? -1 : 0);
}
/*}}}*/
void FS_closefiles(void) /*{{{*/
{
  int i;

  for (i=0; i<capacity; ++i) if (file[i] && !file[i]->dev) FS_close(i);
}
/*}}}*/
