/* Program storage. */
/* #includes */ /*{{{C}}}*//*{{{*/
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "auto.h"
#include "error.h"
#include "fs.h"
#include "program.h"
/*}}}*/

struct Program *Program_new(struct Program *this) /*{{{*/
{
  this->trace=0;
  this->size=0;
  this->numbered=1;
  this->capacity=0;
  this->code=(struct Token**)0;
  this->scope=(struct Scope*)0;
  return this;
}
/*}}}*/
void Program_destroy(struct Program *this) /*{{{*/
{
  while (this->size) Token_destroy(this->code[--this->size]);
  if (this->capacity) free(this->code);
  this->code=(struct Token**)0;
  this->scope=(struct Scope*)0;
}
/*}}}*/
void Program_store(struct Program *this, struct Token *line, long int where) /*{{{*/
{
  int i,last=-1;

  assert(line->type==T_INTEGER || line->type==T_UNNUMBERED);
  if (line->type==T_UNNUMBERED) this->numbered=0;
  if (where)
  {
  for (i=0; i<this->size; ++i)
  {
    assert(this->code[i]->type==T_INTEGER || this->code[i]->type==T_UNNUMBERED);
    if (where>last && where<this->code[i]->u.integer)
    {
      if ((this->size+1)>=this->capacity)
      {
        this->code=realloc(this->code,sizeof(struct Token*)*(this->capacity?(this->capacity*=2):(this->capacity=256)));
      }
      memmove(&this->code[i+1],&this->code[i],(this->size-i)*sizeof(struct Token*));
      this->code[i]=line;
      ++this->size;
      return;
    }
    else if (where==this->code[i]->u.integer)
    {
      Token_destroy(this->code[i]);
      this->code[i]=line;
      return;
    }
    last=this->code[i]->u.integer;
  }
  }
  else i=this->size;
  if ((this->size+1)>=this->capacity)
  {
    this->code=realloc(this->code,sizeof(struct Token*)*(this->capacity?(this->capacity*=2):(this->capacity=256)));
  }
  this->code[i]=line;
  ++this->size;
}
/*}}}*/
int Program_delete(struct Program *this, long int line) /*{{{*/
{
  int i;

  for (i=0; i<this->size; ++i)
  {
    assert(this->code[i]->type==T_INTEGER);
    if (line==this->code[i]->u.integer)
    {
      Token_destroy(this->code[i]);
      if (i+1!=this->size) memmove(&this->code[i],&this->code[i+1],(this->size-i-1)*sizeof(struct Token*));
      --this->size;
      return 0;
    }
  }
  return -1;
}
/*}}}*/
void Program_addScope(struct Program *this, struct Scope *scope) /*{{{*/
{
  struct Scope *s;

  s=this->scope;
  this->scope=scope;
  scope->next=s;
}
/*}}}*/
struct Pc *Program_goLine(struct Program *this, long int line, struct Pc *pc) /*{{{*/
{
  int i;

  for (i=0; i<this->size; ++i)
  {
    if (this->code[i]->type==T_INTEGER && line==this->code[i]->u.integer)
    {
      pc->line=i;
      pc->token=this->code[i]+1;
      return pc;
    }
  }
  return (struct Pc*)0;
}
/*}}}*/
int Program_scopeCheck(struct Program *this, struct Pc *pc, struct Pc *fn) /*{{{*/
{
  struct Scope *scope;

  if (fn==(struct Pc*)0) /* jump from global block must go to global pc */
  {
    for (scope=this->scope; scope; scope=scope->next)
    {
      if (pc->line<scope->begin.line) continue;
      if (pc->line==scope->begin.line && pc->token<=scope->begin.token) continue;
      if (pc->line>scope->end.line) continue;
      if (pc->line==scope->end.line && pc->token>scope->end.token) continue;
      return -1;
    }
  }
  else /* jump from local block must go to local block */
  {
    scope=&(fn->token+1)->u.identifier->sym->u.sub.u.def.scope;
    if (pc->line<scope->begin.line) return -1;
    if (pc->line==scope->begin.line && pc->token<=scope->begin.token) return -1;
    if (pc->line>scope->end.line) return -1;
    if (pc->line==scope->end.line && pc->token>scope->end.token) return -1;
  }
  return 0;
}
/*}}}*/
struct Pc *Program_dataLine(struct Program *this, long int line, struct Pc *pc) /*{{{*/
{
  if ((pc=Program_goLine(this,line,pc))==(struct Pc*)0) return (struct Pc*)0;
  while (pc->token->type!=T_DATA)
  {
    if (pc->token->type==T_EOL) return (struct Pc*)0;
    else ++pc->token;
  }
  return pc;
}
/*}}}*/
long int Program_lineNumber(struct Program *this, struct Pc *pc) /*{{{*/
{
  if (pc->line==-1) return 0;
  if (this->numbered) return (this->code[pc->line]->u.integer);
  else return (pc->line+1);
}
/*}}}*/
struct Pc *Program_beginning(struct Program *this, struct Pc *pc) /*{{{*/
{
  if (this->size==0) return (struct Pc*)0;
  else
  {
    pc->line=0;
    pc->token=this->code[0]+1;
    return pc;
  }
}
/*}}}*/
struct Pc *Program_end(struct Program *this, struct Pc *pc) /*{{{*/
{
  if (this->size==0) return (struct Pc*)0;
  else
  {
    pc->line=this->size-1;
    pc->token=this->code[this->size-1];
    while (pc->token->type!=T_EOL) ++pc->token;
    return pc;
  }
}
/*}}}*/
struct Pc *Program_nextLine(struct Program *this, struct Pc *pc) /*{{{*/
{
  if (pc->line+1==this->size) return (struct Pc*)0;
  else
  {
    pc->token=this->code[++pc->line]+1;
    return pc;
  }
}
/*}}}*/
int Program_skipEOL(struct Program *this, struct Pc *pc, int dev, int tr) /*{{{*/
{
  if (pc->token->type==T_EOL)
  {
    if (pc->line==-1 || pc->line+1==this->size) return 0;
    {
      pc->token=this->code[++pc->line]+1;
      Program_trace(this,pc,dev,tr);
      return 1;
    }
  }
  else return 1;
}
/*}}}*/
void Program_trace(struct Program *this, struct Pc *pc, int dev, int tr) /*{{{*/
{
  if (tr && this->trace && pc->line!=-1)
  {
    char buf[40];

    sprintf(buf,"<%ld>\n",this->code[pc->line]->u.integer);
    FS_putChars(dev,buf);
  }
}
/*}}}*/
void Program_PCtoError(struct Program *this, struct Pc *pc, struct Value *v) /*{{{*/
{
  struct String s;

  String_new(&s);
  if (pc->line>=0)
  {
    if (pc->line<(this->size-1) || pc->token->type!=T_EOL)
    {
      String_appendPrintf(&s," in line %ld at:\n",Program_lineNumber(this,pc));
      Token_toString(this->code[pc->line],(struct Token*)0,&s,(int*)0,-1);
      Token_toString(this->code[pc->line],pc->token,&s,(int*)0,-1);
      String_appendPrintf(&s,"^\n");
    }
    else
    {
      String_appendPrintf(&s," at: end of program\n");
    }
  }
  else
  {
    String_appendPrintf(&s," at: ");
    if (pc->token->type!=T_EOL) Token_toString(pc->token,(struct Token*)0,&s,(int*)0,-1);
    else String_appendPrintf(&s,"end of line\n");
  }
  Value_errorSuffix(v,s.character);
  String_destroy(&s);
}
/*}}}*/
struct Value *Program_merge(struct Program *this, int dev, struct Value *value) /*{{{*/
{
  struct String s;
  int l,err=0;

  l=0;
  while (String_new(&s),(err=FS_appendToString(dev,&s,'\n',1))!=-1 && s.length)
  {
    struct Token *line;

    ++l;
    if (l!=1 || s.character[0]!='#') 
    {
      line=Token_newCode(s.character);
      if (line->type==T_INTEGER && line->u.integer>0) Program_store(this,line,this->numbered?line->u.integer:0);
      else if (line->type==T_UNNUMBERED) Program_store(this,line,0);
      else
      {
        Token_destroy(line);
        return Value_new_ERROR(value,INVALIDLINE,l);
      }
    }
    String_destroy(&s);
  }
  String_destroy(&s);
  if (err) return Value_new_ERROR(value,IOERROR,FS_errmsg);
  return (struct Value*)0;
}
/*}}}*/
struct Value *Program_list(struct Program *this, int dev, struct Pc *from, struct Pc *to, struct Value *value) /*{{{*/
{
  int i,w=0;
  int indent=0;
  struct String s;

  if (this->size==0) return (struct Value*)0;
  for (i=0; i<this->size; ++i) if (this->code[i]->type==T_INTEGER)
  {
    int nw,ln;
    for (ln=this->code[i]->u.integer,nw=1; ln/=10; ++nw);
    if (nw>w) w=nw;
  }
  for (i=0; i<this->size; ++i)
  {
    String_new(&s);
    Token_toString(this->code[i],(struct Token*)0,&s,&indent,w);
    if ((from==(struct Pc *)0 || from->line<=i) && (to==(struct Pc*)0 || to->line>=i))
    {
      if (FS_putString(dev,&s)==-1) return Value_new_ERROR(value,IOERROR,FS_errmsg);
    }
    String_destroy(&s);
  }
  return (struct Value*)0;
}
/*}}}*/
void Program_renum(struct Program *this, int first, int inc) /*{{{*/
{
  int i;
  struct Token *token;

  for (i=0; i<this->size; ++i)
  {
    for (token=this->code[i]; token->type!=T_EOL; ++token)
    {
      if (token->type==T_GOTO || token->type==T_GOSUB || token->type==T_RESTORE || token->type==T_RESUME)
      {
        ++token;
        while (token->type==T_INTEGER)
        {
          struct Pc dst;

          if (Program_goLine(this,token->u.integer,&dst)) token->u.integer=first+dst.line*inc;
          if ((token+1)->type==T_COMMA) ++token;
          else break;
        }
      }
    }
  }
  for (i=0; i<this->size; ++i)
  {
    assert(this->code[i]->type==T_INTEGER || this->code[i]->type==T_UNNUMBERED);
    this->code[i]->type=T_INTEGER;
    this->code[i]->u.integer=first+i*inc;
  }
  this->numbered=1;
}
/*}}}*/
void Program_unnum(struct Program *this) /*{{{*/
{
  char *ref;
  int i;
  struct Token *token;

  ref=malloc(this->size);
  memset(ref,0,this->size);
  for (i=0; i<this->size; ++i)
  {
    for (token=this->code[i]; token->type!=T_EOL; ++token)
    {
      if (token->type==T_GOTO || token->type==T_GOSUB || token->type==T_RESTORE || token->type==T_RESUME)
      {
        ++token;
        while (token->type==T_INTEGER)
        {
          struct Pc dst;

          if (Program_goLine(this,token->u.integer,&dst)) ref[dst.line]=1;
          if ((token+1)->type==T_COMMA) ++token;
          else break;
        }
      }
    }
  }
  for (i=0; i<this->size; ++i)
  {
    assert(this->code[i]->type==T_INTEGER || this->code[i]->type==T_UNNUMBERED);
    if (!ref[i])
    {
      this->code[i]->type=T_UNNUMBERED;
      this->numbered=0;
    }
  }
  free(ref);
}
/*}}}*/
